/*
  Copyright 2019 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth

import com.kakao.sdk.auth.model.AccessTokenResponse
import com.kakao.sdk.auth.model.AgtResponse
import com.kakao.sdk.auth.model.OAuthToken
import com.kakao.sdk.auth.network.kauth
import com.kakao.sdk.common.network.ApiFactory
import com.kakao.sdk.common.util.KakaoJson
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.model.*
import retrofit2.Call
import retrofit2.Callback
import retrofit2.HttpException
import retrofit2.Response

// TODO: 문서화 하고 싶은것과 안하고 싶은게 공존한다. 분리를 시도해보자.
// TODO: ~ Client 계열 클래스들을 object로 변경해도 좋을지 검토
/**
 * 카카오 OAuth 서버에서 제공하는 API 를 사용하기 위한 클라이언트.
 */
class AuthApiClient(
    private val authApi: AuthApi = ApiFactory.kauth.create(AuthApi::class.java),
    private val tokenManager: TokenManageable = TokenManagerProvider.currentManager
) {

    /**
     * [AuthCodeClient] 를 이용하여 발급 받은 인증 코드를 사용하여 액세스 토큰을 발급.
     *
     * @param authCode 인증 코드
     * @callback [OAuthToken] 반환. 토큰은 현재 설정된 [TokenManageable]에 자동으로 저장.
     */
    @JvmOverloads
    fun issueAccessToken(
        authCode: String,
        redirectUri: String = "kakao${KakaoSdk.applicationContextInfo.appKey}://oauth",
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        androidKeyHash: String = KakaoSdk.applicationContextInfo.signingKeyHash,
        callback: (token: OAuthToken?, error: Throwable?) -> Unit
    ) {
        authApi.issueAccessToken(
            clientId = clientId,
            redirectUri = redirectUri,
            androidKeyHash = androidKeyHash,
            authCode = authCode
        ).enqueue(object: Callback<AccessTokenResponse> {
            override fun onFailure(call: Call<AccessTokenResponse>, t: Throwable) {
                callback(null, t)
            }

            override fun onResponse(
                call: Call<AccessTokenResponse>,
                response: Response<AccessTokenResponse>
            ) {
                if (response.isSuccessful) {
                    response.body()?.let {
                        val token = OAuthToken.fromResponse(it)
                        tokenManager.setToken(token)
                        callback(token, null)
                        return
                    }
                    callback(null, ClientError(ClientErrorCause.Unknown, "No body"))
                } else {
                    callback(null, translateError(HttpException(response)))
                }
            }
        })
    }

    /**
     * 리프레시 토큰으로 액세스 토큰을 갱신.
     *
     * @param oldToken 이전에 발급 받았던 토큰. 기본 값은 [TokenManageable]에 저장된 토큰.
     * @param callback [OAuthToken] 반환. 토큰은 현재 설정된 [TokenManageable]에 자동으로 저장.
     *
     * @throws ClientErrorCause.TokenNotFound
     */
    @JvmOverloads
    fun refreshAccessToken(
        oldToken: OAuthToken = tokenManager.getToken() ?: throw ClientError(ClientErrorCause.TokenNotFound),
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        androidKeyHash: String = KakaoSdk.applicationContextInfo.signingKeyHash,
        callback: (token: OAuthToken?, error: Throwable?) -> Unit) {
        authApi.issueAccessToken(
            clientId = clientId,
            redirectUri = null,
            androidKeyHash = androidKeyHash,
            refreshToken = oldToken.refreshToken,
            grantType = Constants.REFRESH_TOKEN
        ).enqueue(object : Callback<AccessTokenResponse> {
            override fun onFailure(call: Call<AccessTokenResponse>, t: Throwable) {
                callback(null, t)
            }

            override fun onResponse(
                call: Call<AccessTokenResponse>,
                response: Response<AccessTokenResponse>
            ) {
                if (response.isSuccessful) {
                    response.body()?.let {
                        val token = OAuthToken.fromResponse(it, oldToken)
                        tokenManager.setToken(token)
                        callback(token, null)
                        return
                    }
                    callback(null, ClientError(ClientErrorCause.Unknown, "No body"))
                } else {
                    callback(null, translateError(HttpException(response)))
                }
            }

        })
    }

    /**
     * @suppress
     */
    fun agt(
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        callback: (response: AgtResponse?, error: Throwable?) -> Unit
    ) {
        tokenManager.getToken()?.accessToken?.let {
            authApi.agt(clientId, it).enqueue(object: Callback<AgtResponse> {
                override fun onFailure(call: Call<AgtResponse>, t: Throwable) {
                    callback(null, t)
                }

                override fun onResponse(call: Call<AgtResponse>, response: Response<AgtResponse>) {
                    response.body()?.let {
                        callback(it, null)
                        return
                    }
                    try {   // 예외케이스로 서버에서 JSON이 아니라 html을 내려줄 때 fromJson이 던지는 예외처리  TODO 다른 곳들도 대응 필요
                        callback(null, translateError(HttpException(response)))
                    } catch (error: Throwable) {
                        callback(null, ClientError(ClientErrorCause.Unknown, error.toString()))
                    }
                }
            })
            return
        }
        callback(null, ClientError(ClientErrorCause.TokenNotFound, "Access token not found."))
    }

    fun refreshAccessToken(
        oldToken: OAuthToken = tokenManager.getToken() ?: throw ClientError(ClientErrorCause.TokenNotFound),
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        androidKeyHash: String = KakaoSdk.applicationContextInfo.signingKeyHash
    ): OAuthToken {
        val response =
            authApi.issueAccessToken(
                clientId = clientId,
                redirectUri = null,
                androidKeyHash = androidKeyHash,
                refreshToken = oldToken.refreshToken,
                grantType = Constants.REFRESH_TOKEN
            ).execute()

        val token = response.body()?.let {
            OAuthToken.fromResponse(it, oldToken)
        } ?: throw translateError(HttpException(response))

        tokenManager.setToken(token)
        return token
    }

    companion object {
        /**
         * @suppress
         */
        fun translateError(t: Throwable): Throwable {
            if (t is HttpException) {
                val errorString = t.response()?.errorBody()?.string()
                val response =
                    KakaoJson.fromJson<AuthErrorResponse>(errorString!!, AuthErrorResponse::class.java)
                val cause = KakaoJson.fromJson(response.error, AuthErrorCause::class.java)
                    ?: AuthErrorCause.Unknown
                return OAuthError(t.code(), cause, response)
            }
            return t
        }

        /**
         * 간편한 API 호출을 위해 기본 제공되는 singleton 객체
         */
        @JvmStatic
        val instance by lazy { AuthApiClient() }
    }
}