/*
  Copyright 2019 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.user

import com.kakao.sdk.auth.TokenManageable
import com.kakao.sdk.auth.TokenManagerProvider
import com.kakao.sdk.auth.network.kapiWithOAuth
import com.kakao.sdk.common.network.ApiCallback
import com.kakao.sdk.common.network.ApiFactory
import com.kakao.sdk.user.model.AccessTokenInfo
import com.kakao.sdk.user.model.User
import com.kakao.sdk.user.model.UserServiceTerms
import com.kakao.sdk.user.model.UserShippingAddresses
import java.util.*

/**
 * 사용자관리 API 호출을 담당하는 클라이언트.
 */
class UserApiClient (
    private val userApi: UserApi = ApiFactory.kapiWithOAuth.create(UserApi::class.java),
    private val tokenManager: TokenManageable = TokenManagerProvider.currentManager
) {

    /**
     * 사용자 정보 요청.
     */
    @JvmOverloads
    fun me(secureReSource: Boolean = true, callback: (user: User?, error: Throwable?) -> Unit) {
        userApi.me(secureReSource)
            .enqueue(object: ApiCallback<User>() {
                override fun onComplete(model: User?, error: Throwable?) {
                    callback(model, error)
                }
            })
    }

    /**
     * 현재 로그인한 사용자의 엑세스 토큰 정보 보기.
     *
     * [me] 에서 제공되는 다양한 사용자 정보 없이 가볍게 토큰의 유효성을 체크하는 용도로 추천.
     * 액세스 토큰이 만료된 경우 자동으로 갱신된 새로운 액세스 토큰 정보 반환.
     */
    fun accessTokenInfo(callback: (tokenInfo: AccessTokenInfo?, error: Throwable?) -> Unit) {
        userApi.accessTokenInfo()
            .enqueue(object: ApiCallback<AccessTokenInfo>() {
                override fun onComplete(model: AccessTokenInfo?, error: Throwable?) {
                    callback(model, error)
                }
            })
    }

    /**
     * User 클래스에서 제공되고 있는 사용자의 부가정보를 신규저장 및 수정.
     *
     * 저장 가능한 키 이름은 카카오 디벨로퍼스 > 카카오 로그인 > 사용자 프로퍼티 메뉴에서 확인.
     * 앱 연결 시 기본 저장되는 nickname, profile_image, thumbnail_image 값도 덮어쓰기 가능하며 새로운 컬럼을 추가하면 해당 키 이름으로 정보 저장 가능.
     */
    fun updateProfile(properties: Map<String, String>, callback: (error: Throwable?) -> Unit) {
        userApi.updateProfile(properties)
            .enqueue(object: ApiCallback<Unit>() {
                override fun onComplete(model: Unit?, error: Throwable?) {
                    callback(error)
                }
            })
    }

    /**
     * 현재 토큰을 강제로 만료시키고 로그아웃.
     *
     * API 호출 결과와 관계 없이 [TokenManageable]에 저장된 토큰을 자동으로 삭제함.
     */
    fun logout(callback: (error: Throwable?) -> Unit) {
        userApi.logout()
            .enqueue(object: ApiCallback<Unit>() {
                override fun onComplete(model: Unit?, error: Throwable?) {
                    tokenManager.clear()
                    callback(error)
                }
            })
    }

    /**
     * 연결 끊기. 카카오 로그인을 통한 사용자와 서비스 간의 연결 관계를 해제하고 사용자의 정보 제공 및 카카오 플랫폼 사용을 중단.
     *
     * API 호출에 성공하면 [TokenManageable]에 저장된 토큰을 자동으로 삭제함.
     */
    fun unlink(callback: (error: Throwable?) -> Unit) {
        userApi.unlink()
            .enqueue(object: ApiCallback<Unit>() {
                override fun onComplete(model: Unit?, error: Throwable?) {
                    if (error == null) {
                        tokenManager.clear()
                    }
                    callback(error)
                }
            })
    }

    /**
     * 사용자의 배송지 정보 획득.
     *
     * @param fromUpdateAt
     * @param pageSize
     */
    @JvmOverloads
    fun shippingAddresses(
        fromUpdateAt: Date? = null,
        pageSize: Int? = null,
        callback: (userShippingAddresses: UserShippingAddresses?, error: Throwable?) -> Unit
    ) {
        userApi.shippingAddresses(fromUpdateAt = fromUpdateAt, pageSize = pageSize)
            .enqueue(object: ApiCallback<UserShippingAddresses>() {
                override fun onComplete(model: UserShippingAddresses?, error: Throwable?) {
                    callback(model, error)
                }
            })
    }

    /**
     * 사용자의 배송지 정보 획득.
     *
     * @param addressId 가져올 배송지 id
     */
    fun shippingAddresses(
        addressId: Long,
        callback: (userShippingAddresses: UserShippingAddresses?, error: Throwable?) -> Unit
    ) {
        userApi.shippingAddresses(addressId)
            .enqueue(object: ApiCallback<UserShippingAddresses>() {
                override fun onComplete(model: UserShippingAddresses?, error: Throwable?) {
                    callback(model, error)
                }
            })
    }

    /**
     * 사용자가 카카오 간편가입을 통해 동의한 서비스 약관 내역 반환.
     */
    fun serviceTerms(callback: (userServiceTerms: UserServiceTerms?, error: Throwable?) -> Unit) {
        userApi.serviceTerms()
            .enqueue(object: ApiCallback<UserServiceTerms>() {
                override fun onComplete(model: UserServiceTerms?, error: Throwable?) {
                    callback(model, error)
                }
            })
    }

    companion object {
        /**
         * 간편한 API 호출을 위해 기본 제공되는 singleton 객체
         */
        @JvmStatic
        val instance by lazy { UserApiClient() }
    }
}
